package algorithms;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Collections;

import geometry.CompareVertexByY;
import geometry.Edge;
import geometry.Geometry;
import geometry.Polygon;
import geometry.Vertex;
import grafico.InterfaceGrafica;

public class MakeMonotone implements Runnable {
	private ArrayList<Edge> EdgeList;
	private InterfaceGrafica ig;
	private Polygon poly;
	
	
	public MakeMonotone(InterfaceGrafica ig, Polygon poly) {
		this.ig = ig;
		this.poly = poly;
	}

	@SuppressWarnings("unchecked")
	private void makeMonotone() {
		
		ig.setRunning(true);
		
		EdgeList = new ArrayList<Edge>();
		
		poly.makeEdgeList();
		
		ig.pintaPoligono(poly, 0);
		ig.copiaTela(0, 1);
		
		Geometry.initializeVertexStatus(poly);
		
		ArrayList<Vertex> events = poly.getVertexList();
		
		for(Vertex v : events){
			System.out.println("(" + v.getPoint().getX() + ", " + v.getPoint().getY() + ")");
		}
		
		Collections.sort(events, new CompareVertexByY());
		Collections.reverse(events);
		
		System.out.println("Starting...");
		for(Vertex v : events){
//			System.out.print("(" + v.getPoint().getX() + ", " + v.getPoint().getY() + ")");
//			System.out.println(" type: " + v.getStatus());
//			for(Edge e : EdgeList){
//				System.out.print("(" + e.getEndPoint1().getPoint().getX() + ", " + e.getEndPoint1().getPoint().getY() + ")");
//				System.out.println(" - (" + e.getEndPoint2().getPoint().getX() + ", " + e.getEndPoint2().getPoint().getY() + ")");
//			}
			ig.copiaTela(1, 2);
			ig.pintaPonto(v.getPoint(), Color.red, 2);
			ig.pintaSegmentoDeRetaHorizontal(v.getPoint().getY(), Color.red, 2);
			ig.pintaTela(2);
			if(v.getStatus() == Geometry.START){
				handleStartVertex(v);
			} else if(v.getStatus() == Geometry.END){
				handleEndVertex(v);
			} else if(v.getStatus() == Geometry.SPLIT){
				handleSplitVertex(v);
			} else if(v.getStatus() == Geometry.MERGE){
				handleMergeVertex(v);
			} else { // REGULAR
				handleRegularVertex(v);
			}
			ig.pintaTelaSemDelay(1);
		}
		
		ig.setRunning(false);
	}
	
	// Returns the edge in the sweep line directly left of vertex v
	Edge directlyLeftOf(Vertex v){
		Edge resp = null;
		for(Edge e : EdgeList){
			if(Geometry.left(e.getEndPoint1().getPoint(), e.getEndPoint2().getPoint(), v.getPoint())){
				if(resp == null || Geometry.segmentIsLeftOf(resp, e)){
					resp = e;					
				}
			}
		}
		if(resp == null){
			System.err.println("No segment left of vertex");
			return null;
		}
		return resp;
	}
	
	private void handleRegularVertex(Vertex v) {
		if(v.getVNext().getPoint().getY() < v.getPoint().getY()){
			if(v.getEPrev().getHelper().getStatus() == Geometry.MERGE){
				processDiagonal(v, v.getEPrev().getHelper());
			}
			EdgeList.remove(v.getEPrev());
			v.getENext().setHelper(v);
			EdgeList.add(v.getENext());
		} else {
			Edge e = directlyLeftOf(v);
			if(e.getHelper().getStatus() == Geometry.MERGE){
				processDiagonal(v, e.getHelper());
			}
			e.setHelper(v);			
		}
	}

	private void handleMergeVertex(Vertex v){
		if(v.getEPrev().getHelper().getStatus() == Geometry.MERGE){
			processDiagonal(v, v.getEPrev().getHelper());
		}
		EdgeList.remove(v.getEPrev());
		Edge e = directlyLeftOf(v);
		if(e.getHelper().getStatus() == Geometry.MERGE){
			// TODO - Why is this diagonal necessary?
			processDiagonal(v, e.getHelper());
		}
		e.setHelper(v);
	}

	private void handleSplitVertex(Vertex v){
		Edge e = directlyLeftOf(v);
		processDiagonal(v, e.getHelper());
		e.setHelper(v);
		v.getENext().setHelper(v);
		EdgeList.add(v.getENext());
	}

	private void handleEndVertex(Vertex v){
		if(v.getEPrev().getHelper().getStatus() == Geometry.MERGE){
			processDiagonal(v, v.getEPrev().getHelper());
		}
		EdgeList.remove(v.getEPrev());
	}

	private void handleStartVertex(Vertex v) {
		v.getENext().setHelper(v);
		EdgeList.add(v.getENext());
	}
	
	private void processDiagonal(Vertex v1, Vertex v2) {
		ig.pintaSegmentoDeReta(v1.getPoint(), v2.getPoint(), Color.red, 1);
		ig.pintaTela(1);
	}

	public void run() {
		makeMonotone();		
	}

}
