package algorithms;

import geometry.Ponto;
import geometry.CompareByX;
import geometry.CompareByY;
import geometry.iPoint;
import grafico.InterfaceGrafica;

import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Vector;

import javax.swing.JButton;


public class MinDist implements Runnable {
	
	private static final int CANVAS_H = 540;
	public static int LARGURA = 8;
	private InterfaceGrafica interfaceGrafica;
	private Vector<Ponto> pontos;
	private Vector<Ponto> pontosPorX;
	private Vector<Ponto> pontosPorY;
	private boolean rodaIngenuo = false;
	private boolean rodaDivConq = false;
	private Ponto min1, min2;
	double minDist;
	int count;

	public MinDist(InterfaceGrafica interfaceGrafica, Vector<Ponto> pontos) {
		pontos = new Vector<Ponto>();
		
		setInterfaceGrafica(interfaceGrafica);
		setPontos(pontos);
	}
	
	public MinDist(InterfaceGrafica interfaceGrafica, ArrayList<iPoint> ipontos) {
		pontos = new Vector<Ponto>();
		
		setInterfaceGrafica(interfaceGrafica);
		
		for(iPoint p : ipontos){
			pontos.add( new Ponto( (double)p.getX(), (double)p.getY() ) );
		}
	}

	
	public void run() {
		interfaceGrafica.setRunning(true);
		if(rodaIngenuo){
			ingenuo();
			rodaIngenuo = false;
		}
		else if(rodaDivConq){
			nlogn();
		}
		interfaceGrafica.setRunning(false);
	}
	
	public int ingenuo(){
		
		minDist = Double.MAX_VALUE;
		count = 0;
		min1 = null;
		min2 = null;
		
		if(getPontos() == null) return 0;
		
		interfaceGrafica.copiaTela(0, 1);
		
		for(int i = 0;i < getPontos().size();i++){
			Ponto pontoAtual = getPontos().get(i);
			interfaceGrafica.pintaPonto(pontoAtual, Color.green, 1);
			
			for(int j = i+1;j < getPontos().size();j++){
				interfaceGrafica.pintaPonto(pontoAtual, Color.green, 1);
				count++;
				if(minDist > pontoAtual.dist(getPontos().get(j))){
					minDist = pontoAtual.dist(getPontos().get(j));
					min1 = pontoAtual;
					min2 = getPontos().get(j);
				}
				interfaceGrafica.pintaSegmentoDeReta(pontoAtual, getPontos().get(j), 1);
				interfaceGrafica.pintaSegmentoDeReta(pontoAtual, getPontos().get(j), 2);
				interfaceGrafica.pintaTela(1);
				interfaceGrafica.copiaTela(0, 1);
				interfaceGrafica.pintaPonto(pontoAtual, Color.green, 1);
				interfaceGrafica.pintaTela(1);
			}
			interfaceGrafica.pintaPonto(pontoAtual, 1);
		}
		pintaResposta();
		interfaceGrafica.setPassosIngenuo(count);
		return count;
	}


	private void pintaResposta() {
		interfaceGrafica.limpaTela(1);
		
		if(interfaceGrafica.isApaga()){
			interfaceGrafica.copiaTela(0, 1);
		} else {
			interfaceGrafica.copiaTela(2, 1);
		}
		interfaceGrafica.pintaPonto(min1, Color.green, 1);
		interfaceGrafica.pintaPonto(min2, Color.green, 1);
		interfaceGrafica.pintaSegmentoDeReta(min1, min2, Color.red, 1);
		interfaceGrafica.pintaTela(1);
	}

//	public void actionPerformed(ActionEvent e) {
//		JButton source = (JButton) e.getSource();
//		if(source.getText().equals("Força Bruta:")){
//			interfaceGrafica.setPassosIngenuo(-1);
//			MinDist mindist = new MinDist(interfaceGrafica);
//			mindist.setPontos(getPontos());
//			
//			mindist.setRodaIngenuo(true);
//			mindist.start();
//		} else {
//			interfaceGrafica.setPassosNLogN(-1);
//			MinDist mindist = new MinDist(interfaceGrafica);
//			mindist.setPontos(getPontos());
//			
//			mindist.setRodaDivConq(true);
//			mindist.start();
//		}
//	}

	public void setRodaIngenuo(boolean roda) {
		rodaIngenuo = roda;
		rodaDivConq = !roda;
	}
	
	public void setRodaDivConq(boolean roda) {
		rodaDivConq = roda;
		rodaIngenuo = !roda;
	}

	public synchronized void setPontos(Vector<Ponto> pontos) {
		this.pontos = pontos;
	}

	public synchronized Vector<Ponto> getPontos() {
		return pontos;
	}

	private synchronized void setInterfaceGrafica(InterfaceGrafica interfaceGrafica) {
		this.interfaceGrafica = interfaceGrafica;
	}
	
	
	@SuppressWarnings("unchecked")
	private double divideAndConquer(int left, int right, int tela){
		
		if(left >= right)
			return 1e18;
		
		interfaceGrafica.copiaTela(tela - 1, tela);
		interfaceGrafica.copiaTela(tela - 1, 1);
		
		if(left == right - 1){
			count++;
			interfaceGrafica.pintaSegmentoDeReta(pontosPorX.get(left), pontosPorX.get(right), 1);
			interfaceGrafica.pintaSegmentoDeReta(pontosPorX.get(left), pontosPorX.get(right), 2);
			interfaceGrafica.pintaTela(1);
			interfaceGrafica.pintaTela(tela - 1);
			double dist = pontosPorX.get(left).dist(pontosPorX.get(right));
			
			if(dist < minDist){
        		minDist = dist;
        		min1 = pontosPorX.get(left);
        		min2 = pontosPorX.get(right);
        	}
			
			return dist;			
		}
		
		int mid = (left + right)/2;
		
		interfaceGrafica.pintaSegmentoDeReta(new Ponto(pontosPorX.get(mid).getX(), 0),
											 new Ponto(pontosPorX.get(mid).getX(), CANVAS_H), Color.red, 1);
		interfaceGrafica.pintaSegmentoDeReta(new Ponto(pontosPorX.get(mid).getX(), 0),
				 							 new Ponto(pontosPorX.get(mid).getX(), CANVAS_H), Color.red, tela);
		interfaceGrafica.pintaTela(tela);
		
	    // double best = Math.min(divideAndConquer(left, mid, tela + 1), divideAndConquer(mid + 1, right, tela + 1));
		double best = divideAndConquer(left, mid, tela + 1);
		interfaceGrafica.pintaTela(tela);
		
		best = Math.min(best, divideAndConquer(mid + 1, right, tela + 1));
	    count++;
	    
	    interfaceGrafica.pintaSegmentoDeReta(new Ponto(pontosPorX.get(mid).getX(), 0),
				 							 new Ponto(pontosPorX.get(mid).getX(), CANVAS_H), Color.blue, tela);
	    
	    interfaceGrafica.pintaTela(tela);
	    
	    pintaFaixa(mid, best, tela);

	    Vector<Ponto> faixa = new Vector<Ponto>();
	    // Filtra os pontos dentro da faixa
	    for(int i = left; i <= right; i++){
	    	count++;
	        if(pontosPorX.get(i).getX() >= pontosPorX.get(mid).getX() - best
	        		&& pontosPorX.get(i).getX() <= pontosPorX.get(mid).getX() + best)
	        	faixa.add(new Ponto(pontosPorX.get(i).getX(), pontosPorX.get(i).getY()));
	    	}
	    
	    Collections.sort(faixa, new CompareByY());
	    
	    pintaPontosDoConjunto(faixa, tela);
	    interfaceGrafica.copiaTela(tela, 1);

	    // Para cada ponto, olha apenas para os 6 vizinhos acima
	    double minDistLocal = best;
	    for(int i = 0; i < faixa.size(); i++){
	    	interfaceGrafica.copiaTela(tela, 1);
	    	interfaceGrafica.pintaTela(1);
	    	//interfaceGrafica.pintaPontoColorido(faixa.get(i), Color.green, 1);
	    	for(int j = i + 1; j < i + 7 && j < faixa.size(); j++){
	        //for(int j = i + 1; j < faixa.size() && j < faixa.size() && Math.abs(faixa.get(i).getY() - faixa.get(j).getY()) < minDistLocal; j++){
	        	count++;
	        	interfaceGrafica.pintaSegmentoDeReta(faixa.get(i), faixa.get(j), 1);
	        	interfaceGrafica.pintaSegmentoDeReta(faixa.get(i), faixa.get(j), 2);
	        	interfaceGrafica.pintaTela(1);
	        	
	        	double dist = faixa.get(i).dist(faixa.get(j));
	            minDistLocal = Math.min(minDistLocal, dist);
	            if(dist < minDist){
            		minDist = dist;
            		min1 = faixa.get(i);
            		min2 = faixa.get(j);
            	}
	        }
	    }

	    interfaceGrafica.pintaTela(tela - 1);
	    
	    return minDistLocal;
	}
	
	private void pintaPontosDoConjunto(Vector<Ponto> faixa, int tela) {
		for(Ponto p : faixa)
			interfaceGrafica.pintaPonto(p, Color.blue, tela);
		interfaceGrafica.pintaTela(tela);
	}


	private void pintaFaixa(int mid, double best, int tela) {
		double faixa = pontosPorX.get(mid).getX();
		double faixaLeft = Math.max(0, faixa - best);
		double faixaRight = Math.min(InterfaceGrafica.CANVAS_W, faixa + best);
		
		interfaceGrafica.pintaSegmentoDeReta(new Ponto(faixaLeft, 0),
											 new Ponto(faixaLeft, CANVAS_H), tela);

		interfaceGrafica.pintaSegmentoDeReta(new Ponto(faixaRight, 0),
											 new Ponto(faixaRight, CANVAS_H), tela);
		
		interfaceGrafica.pintaTela(tela);
	}
	
	@SuppressWarnings("unchecked")
	public int nlogn(){
		count = 0;
		min1 = null;
		min2 = null;
		minDist = Double.MAX_VALUE;
		
		if(getPontos() == null) return 0;
		
		pontosPorX = (Vector<Ponto>)pontos.clone();
		pontosPorY = (Vector<Ponto>)pontos.clone();
		
		// Ordena por x
		Collections.sort(pontosPorX, new CompareByX());
		
		// Ordena por y
		Collections.sort(pontosPorY, new CompareByY());
		
		interfaceGrafica.copiaTela(0, 1);
		interfaceGrafica.copiaTela(0, 2);
		double dist = divideAndConquer(0, getPontos().size() - 1, 3);
		
		pintaResposta();
		interfaceGrafica.setPassosNLogN(count);
		
		return (int) dist;
	}

}
