package algorithms;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Stack;

import geometry.CompareVertexByY;
import geometry.Geometry;
import geometry.Polygon;
import geometry.Vertex;
import grafico.InterfaceGrafica;

public class TriangulateMonotonePolygon implements Runnable {

	private InterfaceGrafica ig;
	private Polygon poly;
	
	public TriangulateMonotonePolygon(InterfaceGrafica ig, Polygon poly) {
		this.ig = ig;
		this.poly = poly;
	}
	
	@SuppressWarnings("unchecked")
	private void triangulate(){
		ig.setRunning(true);
		
		ig.pintaPoligono(poly, 0);
		ig.copiaTela(0, 1);
		
		if(!poly.isMonotone()){
			System.err.println("Polígono não é monótono");
			ig.setRunning(false);
			return;
		}
		
		Geometry.initializeChain(poly);
		
		ArrayList<Vertex> events = poly.getVertexList();
		
		// TODO - Merge O(n)
		Collections.sort(events, new CompareVertexByY());
		Collections.reverse(events);
		
		Stack<Vertex> S = new Stack<Vertex>();
		
		S.push(events.get(0));
		S.push(events.get(1));
		
		events.get(0).setChain(events.get(1).getChain());
		
		for(int i = 2; i < events.size() - 1; i++){
			Vertex v = events.get(i);
			ig.copiaTela(1, 2);
			ig.pintaPonto(v.getPoint(), Color.red, 2);
			ig.pintaTela(2);
			if(v.getChain() != S.peek().getChain()){ // Different chains
				
				// Draw diagonals from v to all vertices on the stack except the last one
				while(S.size() > 1){
					Vertex top = S.pop();
					ig.pintaSegmentoDeReta(v.getPoint(), top.getPoint(), Color.red, 1);
					ig.pintaSegmentoDeReta(v.getPoint(), top.getPoint(), Color.red, 2);
					ig.pintaTela(2);
				}
				S.pop();
				S.add(events.get(i - 1));
				S.add(v);
			} else { // Same chain
				Vertex prev = S.pop();
				
				// While we can draw diagonals from v to the vertices on the stack (all on the same chain)
				while(S.size() > 0 
						&& ((v.getChain() == 0 && Geometry.left(v.getPoint(), prev.getPoint(), S.peek().getPoint()))
						      || (v.getChain() == 1 && !Geometry.left(v.getPoint(), prev.getPoint(), S.peek().getPoint())))){
					prev = S.pop();
					ig.pintaSegmentoDeReta(v.getPoint(), prev.getPoint(), Color.red, 1);
					ig.pintaSegmentoDeReta(v.getPoint(), prev.getPoint(), Color.red, 2);
					ig.pintaTela(2);
				}
				S.push(prev);
				S.push(v);
			}
			ig.pintaTela(1);
		}
		
		// Draw diagonals from the last vertex to all vertices except for the first and last one
		S.pop();
		while(S.size() > 1){
			ig.pintaSegmentoDeReta(events.get(events.size() - 1).getPoint(), S.pop().getPoint(), Color.red, 1);
			ig.pintaTela(1);
		}
		
		ig.setRunning(false);
	}
	
	public void run() {
		triangulate();		
	}
	
	
}
