package algorithms;

import geometry.Geometry;
import geometry.Polygon;
import geometry.Vertex;
import grafico.InterfaceGrafica;

import java.awt.Color;

public class TriangulatePolygon implements Runnable {

	private InterfaceGrafica ig;
	private Polygon poly;
	
	public TriangulatePolygon(InterfaceGrafica ig, Polygon poly) {
		this.ig = ig;
		this.poly = poly;
	}
	
	public void triangulateEarRemoval(){
		
		ig.setRunning(true);
		
		ig.pintaPoligono(poly, 0);
		ig.copiaTela(0, 1);
		
		Geometry.initializeEarStatus(poly);
		paintEars(poly);
		
		for(int i = 0; i < poly.getSize() - 3; i++){
			
			Vertex v = poly.getHead();
			do {
				
				if(v.isEar()){					
					processDiagonal(v.getVPrev(), v.getVNext());
					
					Geometry.removeEar(poly, v);
					
					v.getVNext().setEar(Geometry.isEar(poly, v.getVNext()));
					v.getVPrev().setEar(Geometry.isEar(poly, v.getVPrev()));
					paintEar(poly, v.getVNext(), 1);
					paintEar(poly, v.getVPrev(), 1);
					
					break;
				}
				v = v.getVNext();
			} while(v != poly.getHead());
		}
		
		ig.setRunning(false);
		
	}

	private void paintEar(Polygon poly, Vertex v, int tela){
		ig.pintaPonto(v.getPoint(), Color.blue, 1);
		ig.pintaTela(1);
		if(v.isEar()){
			ig.pintaPonto(v.getPoint(), Color.red, 1);
		} else {
			ig.pintaPonto(v.getPoint(), Color.black, 1);
		}
		ig.pintaTela(1);
	}
	
	private void paintEars(Polygon poly) {
		Vertex v = poly.getHead();
		do {
			paintEar(poly, v, 1);
			
			v = v.getVNext();
		} while(v != poly.getHead());
	}

	private void processDiagonal(Vertex v1, Vertex v2) {
		ig.pintaSegmentoDeReta(v1.getPoint(), v2.getPoint(), Color.red, 1);
		ig.pintaTela(1);
	}

	public void run() {
		triangulateEarRemoval();
	}

	
}
