package geometry;

import java.util.ArrayList;

public class QuadNode {

	private Ponto upperLeftCorner; 
	private Ponto upperRightCorner;
	private Ponto lowerLeftCorner;
	private Ponto lowerRightCorner;
	
	private boolean leaf = true;
	
	private double side;
	
	ArrayList<Ponto> points;
	
	// 0 | 1
	// --+--
	// 2 | 3
	QuadNode children[]; 
	
	QuadNode(Ponto ulc, Ponto lrc){
		
		points = new ArrayList<Ponto>();
		children = new QuadNode[4];
		
		setUpperLeftCorner(ulc);
		setLowerRightCorner(lrc);
		
		setSide(upperLeftCorner.getX() - lowerRightCorner.getX());
		
	}
	
	public void setUpperLeftCorner(Ponto upperLeftCorner) {
		this.upperLeftCorner = upperLeftCorner;
	}
	
	public Ponto getUpperLeftCorner() {
		return upperLeftCorner;
	}
	
	public void setUpperRightCorner(Ponto upperRightCorner) {
		this.upperRightCorner = upperRightCorner;
	}
	
	public Ponto getUpperRightCorner() {
		return upperRightCorner;
	}
	
	public void setLowerLeftCorner(Ponto lowerLeftCorner) {
		this.lowerLeftCorner = lowerLeftCorner;
	}
	
	public Ponto getLowerLeftCorner() {
		return lowerLeftCorner;
	}
	
	public void setLowerRightCorner(Ponto lowerRightCorner) {
		this.lowerRightCorner = lowerRightCorner;
	}
	
	public Ponto getLowerRightCorner() {
		return lowerRightCorner;
	}

	public void setSide(double side) {
		this.side = side;
	}

	public double getSide() {
		return side;
	}

	public void setLeaf(boolean leaf) {
		this.leaf = leaf;
	}

	public boolean isLeaf() {
		return leaf;
	}
	
	Ponto getCenter(){
		return new Ponto(upperLeftCorner.getX() + side/2, lowerLeftCorner.getY() + side/2);
	}
	
	public boolean containsPoint(Ponto p){
		return (p.getX() >= lowerLeftCorner.getX() && p.getX() <= lowerRightCorner.getX()) &&
			   (p.getY() <= upperLeftCorner.getY() && p.getY() >= lowerLeftCorner.getY());
	}
	
	public void addPoint(Ponto p){
		points.add(p);
	}
	
	void splitNode(){
		setLeaf(false);
		
		children[0] = new QuadNode(upperLeftCorner, getCenter());
		children[1] = new QuadNode(new Ponto(upperLeftCorner.getX() + side/2, upperLeftCorner.getY()), 
								   new Ponto(lowerRightCorner.getX(), lowerRightCorner.getY() + side/2));
		children[2] = new QuadNode(new Ponto(upperLeftCorner.getX(), lowerLeftCorner.getY() + side/2), 
				   				   new Ponto(lowerRightCorner.getX() + side/2, lowerRightCorner.getY()));
		children[3] = new QuadNode(getCenter(), lowerRightCorner);
		
		for(Ponto p : points){
			for(int i = 0; i < 4; i++){
				if(children[i].containsPoint(p)){
					children[i].addPoint(p);
					break;
				}
			}
		}		
	}

}
