package grafico;

import geometry.LineSegment;
import geometry.Polygon;
import geometry.Ponto;
import geometry.Vertex;
import geometry.iPoint;

import java.awt.BorderLayout;
import java.awt.Canvas;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Graphics;
import java.awt.GridLayout;
import java.awt.Image;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.ArrayList;
import java.util.Hashtable;

import javax.swing.BorderFactory;
import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JSlider;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import principal.EscritaDeDados;
import principal.LeituraDeDados;

import algorithms.LegalTriangulation;
import algorithms.MakeMonotone;
import algorithms.MinDist;
import algorithms.PlanarPointSetTriangulation;
import algorithms.TriangulateMonotonePolygon;
import algorithms.TriangulatePolygon;


@SuppressWarnings("serial")
public class InterfaceGrafica extends JFrame  implements ChangeListener, ActionListener{

	static final int NUMERO_TELAS = 20;
	static final int LARGURA_PONTO = 8;
	static final int FPS_MIN = 1;
	static final int FPS_MAX = 500;
	static final int FPS_INIT = 15;
	static final public int CANVAS_H = 600;
	static final public int CANVAS_W = 800;

	private JSlider framesPerSecond;
	private Canvas canvas;
	private ArrayList<Image> telas;
	private ArrayList<iPoint> pontos;
	private int delay;
	private boolean frozen = false;
	private boolean go = false;
	
	// Input is a polygon or a planar point set?
	// 0 - Polygon
	// 1 - PPST
	private int option = 0;

	private JMenuBar menuBar;
	
	private JButton drawPolygon;
	private JButton drawPoints;
	private JButton next;
	private JButton apagar;
	private JButton abrir;
	private JButton salvar;
	
	// Algorithm Panel
	private JPanel algos;
	
	// Algorithm Buttons
	private JButton triangulatePolygon;
	private JButton makeMonotone;
	private JButton legalTriangulation;
	private JButton PPST;
	private JButton triangulateMonotone;
	
	private JButton closestBruteForce;
	private JButton closestDivideConquer;
	
	private JCheckBox erase;
	
	private JFileChooser filechooser;
	
	private IngenuoLBL ingenuoLBL;
	private NLognLBL nLognLBL;
	
	private int telaAtual = 0;
	private boolean apaga = true;
	private boolean drawingPolygon = false;
	private boolean drawingPoints = false;
	
	private boolean running = false;

	public InterfaceGrafica() {

		telas = new ArrayList<Image>();
		
		createMenu();
		setFPSBar();	
		setCanvas();
		createButtons();
		
		pontos = new ArrayList<iPoint>();
		
		ingenuoLBL = new IngenuoLBL();
		nLognLBL = new NLognLBL();
		
		filechooser = new JFileChooser();
		
		erase = new JCheckBox("Manter comparações?");
		
		erase.addItemListener(new ItemListener(){

			public void itemStateChanged(ItemEvent e) {
				if(apaga)
					setApaga(false);
				else
					setApaga(true);
				
			}
			
		});
		
		setVisual();
	}
	
	private void createMenu(){
		menuBar = new JMenuBar();
		
		JMenu file = new JMenu("File");
		
		JMenuItem saveItem = new JMenuItem("Save");
		JMenuItem loadItem = new JMenuItem("Load");
		file.add(saveItem);
		file.add(loadItem);
		
		JMenu tools = new JMenu("Tools");
		ButtonGroup rbGroup = new ButtonGroup();
		JRadioButton closestPair = new JRadioButton("Closest Pair");
		closestPair.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				option = 1;
				setCLosestPairPanel();
			}
		});
		
		JRadioButton triangulation = new JRadioButton("Triangulation");
		triangulation.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				option = 0;
				setTriangulationPanel();
			}
		});
		
		JRadioButton radioPPST = new JRadioButton("Planar Triangulation");
		radioPPST.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				option = 1;
				setPPSTPanel();
			}
		});
		
		rbGroup.add(closestPair);
		rbGroup.add(triangulation);
		rbGroup.add(radioPPST);
		tools.add(closestPair);
		tools.add(triangulation);
		tools.add(radioPPST);
				
		menuBar.add(file);
		menuBar.add(tools);
	}

	public synchronized void dorme(){
		if(!frozen){
			try {
				Thread.sleep(delay);
			} catch (InterruptedException e) {}
		} else {
			while(!go){
				try {
					wait();
				} catch (InterruptedException e) {}
			}
			go = false;
		}
	}
	
	private void setCanvas() {

		canvas = new Canvas(){
			public void paint(Graphics g){
				g.drawImage(telas.get(getTelaAtual()), 0, 0, this);
			}
			public void update(Graphics g){
				paint(g);
			}
		};

		canvas.setPreferredSize(new Dimension(CANVAS_W,CANVAS_H));
		
		canvas.addMouseListener(new MouseAdapter() {
			public void mouseClicked(MouseEvent me){
				System.out.println("x: " + me.getX() + " y: " + me.getY());
				if(!isRunning()){
					pontos.add(new iPoint(me.getX(), me.getY()));
					pintaPonto(me.getX(), me.getY(), Color.black, 0);
					if(option == 0){
						if(pontos.size() > 1){
							pintaSegmentoDeReta(pontos.get(pontos.size() - 1), pontos.get(pontos.size() - 2), 0);
						}
					}
					pintaTelaSemDelay(0);
				}
			}
		});
	}
	
	public int getTelaAtual() {
		return telaAtual;
	}
	
	public void setTelaAtual(int tela) {
		this.telaAtual = tela;
	}
	
	public void inicializaImagens(){
		for(int i = 0; i < NUMERO_TELAS; i++){
			Image tmp = createImage(CANVAS_W, CANVAS_H);
			telas.add(tmp);
		}
	}

	private void createButtons() {
		next = new JButton("Step");
		next.addActionListener(this);
		
		apagar = new JButton("Erase");
		apagar.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent e) {
				clearCanvas();
			}
			
		});
		
		abrir = new JButton("Abrir Arquivo");
		abrir.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent e) {
				int returnVal = filechooser.showOpenDialog(InterfaceGrafica.this);
				
				if(returnVal == JFileChooser.APPROVE_OPTION){
					LeituraDeDados.le(filechooser.getSelectedFile());
				}
				else{
					System.err.println("Nao consegui abrir o arquivo");
				}
			}
		});
		
		salvar = new JButton("Salvar Arquivo");
		salvar.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent e) {
				int returnVal = filechooser.showSaveDialog(InterfaceGrafica.this);
				
				if(returnVal == JFileChooser.APPROVE_OPTION){
					EscritaDeDados.escreve(filechooser.getSelectedFile());
				}
				else{
					//nao consegui
				}
			}
		});
		
		//drawPolygon = new JButton("Desenhar Polígono");
		drawPolygon = new JButton("Draw Polygon");
		drawPolygon.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				if(!isDrawingPolygon()){
					drawPolygon.setText("Stop");
					setDrawingPolygon(true);
				} else {
					drawPolygon.setText("Draw Polygon");
					setDrawingPolygon(false);
					if(pontos.size() > 3){
						pintaSegmentoDeReta(pontos.get(0), pontos.get(pontos.size() - 1), 0);
						pintaTelaSemDelay(0);
					}
				}
			}
		});
		
		drawPoints = new JButton("Draw Points");
		drawPoints.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				if(!isDrawingPoints()){
					clearCanvas();
					drawPoints.setText("Stop");
					setDrawingPolygon(true);
				} else {
					drawPolygon.setText("Draw Polygon");
					setDrawingPolygon(false);
					if(pontos.size() > 3){
						pintaSegmentoDeReta(pontos.get(0), pontos.get(pontos.size() - 1), 0);
						pintaTelaSemDelay(0);
					}
				}
			}
		});
		
		closestBruteForce = new JButton("BruteForce");
		closestBruteForce.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent e) {
				
				if(okToRun()){
			
					MinDist MD = new MinDist(InterfaceGrafica.this, pontos);
					MD.setRodaIngenuo(true);
					Thread TMD = new Thread(MD);

					TMD.start();
					
				}
			}
			
		});
		
		closestDivideConquer = new JButton("Divide & Conquer");
		closestDivideConquer.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent e) {
				
				if(okToRun()){
			
					MinDist MD = new MinDist(InterfaceGrafica.this, pontos);
					MD.setRodaDivConq(true);
					Thread TMD = new Thread(MD);

					TMD.start();
					
				}
			}
			
		});

		triangulatePolygon = new JButton("Triangulate");
		triangulatePolygon.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent e) {
				
				if(okToRun()){
			
					TriangulatePolygon TP = new TriangulatePolygon(InterfaceGrafica.this, new Polygon(pontos));
					Thread TTP = new Thread(TP);

					TTP.start();
					
				}
			}
			
		});
		
		makeMonotone = new JButton("Make Monotone");
		makeMonotone.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent e) {
				
				if(okToRun()){
					
					System.out.println("Ok to run");
					MakeMonotone MM = new MakeMonotone(InterfaceGrafica.this, new Polygon(pontos));
					Thread TMM = new Thread(MM);
				
					TMM.start();

				}
			}
			
		});
		
		triangulateMonotone = new JButton("Triangulate Monotone");
		triangulateMonotone.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent e) {
				
				if(okToRun()){
					
					TriangulateMonotonePolygon TMP = new TriangulateMonotonePolygon(InterfaceGrafica.this, new Polygon(pontos));
					Thread TTMP = new Thread(TMP);
				
					TTMP.start();
					
				}
			}
			
		});
		
		legalTriangulation = new JButton("Legal Triangulation");
		legalTriangulation.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent e) {
				if(okToRun()){
					LegalTriangulation LT = new LegalTriangulation(InterfaceGrafica.this, pontos);
					Thread TLT = new Thread(LT);

					TLT.start();
				}
			}
		});
		
		PPST = new JButton("PPST");
		PPST.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent e) {
				if(okToRun()){
					PlanarPointSetTriangulation PST = new PlanarPointSetTriangulation(InterfaceGrafica.this, pontos);
					Thread PSTT = new Thread(PST);

					PSTT.start();
				}
			}
		});
	}

	private void setVisual() {
		this.setSize(600, 600);
		this.setTitle("Triangulation");
		//this.setResizable(false);
		JPanel panel = new JPanel();
		algos = new JPanel(new FlowLayout(FlowLayout.LEFT));
		panel.add(framesPerSecond);
		panel.add(next);
		panel.add(apagar);
		//panel.add(abrir);
		//panel.add(salvar);
		//panel.add(drawPolygon);
		//panel.add(drawPoints);
		//algos.add(ingenuoLBL);
		//algos.add(nLognLBL);
		algos.setBorder(BorderFactory.createLoweredBevelBorder());
		algos.setPreferredSize(new Dimension(150, 10));
		algos.add(triangulatePolygon);
		algos.add(makeMonotone);
		algos.add(triangulateMonotone);
		
		//algos.add(erase);
		this.setJMenuBar(menuBar);
		this.add(canvas, BorderLayout.CENTER);
		this.add(panel, BorderLayout.SOUTH);
		this.add(algos, BorderLayout.WEST);
		
		panel.setBorder(BorderFactory.createLoweredBevelBorder());
		
		this.addWindowListener(new WindowAdapter(){
			public void windowClosing(WindowEvent we){
				System.exit(0);
			}
		});

		this.pack();
	}
	
	private void clearCanvas(){
		pontos.clear();
		limpaTela(0);
		pintaTelaSemDelay(0);
	}
	
	private void setCLosestPairPanel(){
		algos.removeAll();

		algos.add(closestBruteForce);
		algos.add(closestDivideConquer);
		
		this.add(algos, BorderLayout.WEST);
		this.pack();		
	}
	
	private void setTriangulationPanel(){
		algos.removeAll();
		
		algos.add(triangulatePolygon);
		algos.add(makeMonotone);
		algos.add(triangulateMonotone);
		
		this.add(algos, BorderLayout.WEST);
		this.pack();	
	}
	
	private void setPPSTPanel(){
		algos.removeAll();
		
		algos.add(PPST);
		
		this.add(algos, BorderLayout.WEST);
		this.pack();	
	}

	@SuppressWarnings("unchecked")
	private void setFPSBar() {
		framesPerSecond = new JSlider(JSlider.HORIZONTAL, FPS_MIN, FPS_MAX, FPS_INIT);
		framesPerSecond.setMajorTickSpacing(100);
		framesPerSecond.setMinorTickSpacing(25);

		Hashtable labelTable = new Hashtable();
		labelTable.put( new Integer( 1 ), new JLabel("Step") );
		labelTable.put( new Integer( FPS_MAX/10 ), new JLabel("Slow") );
		labelTable.put( new Integer( FPS_MAX ), new JLabel("Fast") );


		framesPerSecond.setLabelTable( labelTable );
		framesPerSecond.setPreferredSize(new Dimension(600,100));

		framesPerSecond.setPaintTicks(true);
		framesPerSecond.setPaintLabels(true);
		framesPerSecond.addChangeListener(this);

		delay = 300;
	}

	public synchronized void actionPerformed(ActionEvent e) {
		if(!go){
			go = true;
			notify();
		}
	}
	
	// Checks if it is ok to run algorithm.
	private boolean okToRun(){
		return (!isRunning() && !isDrawingPolygon());
	}

	public synchronized void stateChanged(ChangeEvent e) {
		JSlider source = (JSlider)e.getSource();
		if (!source.getValueIsAdjusting()) {
			int fps = (int)source.getValue();
			if (fps == FPS_MIN) {
				go = false;
				frozen = true;
			}
			else {
				frozen = false;
			}
			delay = 500-fps;
		}
	}

	public synchronized void passo(){
		if(!frozen){
			try {
				Thread.sleep(delay);
			} catch (InterruptedException e) {
			}
		}
		else{
			while(!go){
				try {
					wait();
				} catch (InterruptedException e) {
				}
			}
			go = false;
		}
	}

	public void limpa() {
		canvas.getGraphics().clearRect(0, 0, CANVAS_W, CANVAS_H);
	}
	
	public void setPassosIngenuo(int num) {
		ingenuoLBL.setPassos(num);
	}
	
	public void setPassosNLogN(int num) {
		nLognLBL.setPassos(num);
	}
	
	public synchronized void addActionListener(ActionListener actionListener){
		ingenuoLBL.addActionListenerIngenuo(actionListener);
		nLognLBL.addActionListenerDivisaoConquista(actionListener);
	}

	public void setApaga(boolean apaga) {
		this.apaga = apaga;
	}
	
	public boolean isApaga() {
		return apaga;
	}

	public void atualiza() {
		canvas.repaint();
	}
	
	public void pintaTela(int tela){
		setTelaAtual(tela);
		atualiza();
		dorme();
	}
	
	public void pintaTelaSemDelay(int tela){
		setTelaAtual(tela);
		atualiza();
	}
	
	public void pintaPonto(int x, int y, Color c, int tela){

		Graphics graphics = telas.get(tela).getGraphics();
		graphics.setColor(c);

		graphics.fillOval(x - LARGURA_PONTO/2, y - LARGURA_PONTO/2, LARGURA_PONTO, LARGURA_PONTO);

	}
	
	public void pintaPonto(int x, int y, int tela){
		pintaPonto(x, y, Color.black, tela);		
	}
	
	public void pintaPonto(iPoint ponto, Color c, int tela){
		pintaPonto(ponto.getX(), ponto.getY(), c, tela);
	}
	
	public void pintaPonto(iPoint ponto, int tela){
		pintaPonto(ponto.getX(), ponto.getY(), Color.black, tela);
	}
	
	public void pintaPonto(Ponto ponto, Color c, int tela){
		pintaPonto((int)ponto.getX(), (int)ponto.getY(), c, tela);
	}
	
	public void pintaPonto(Ponto ponto, int tela){
		pintaPonto((int)ponto.getX(), (int)ponto.getY(), Color.black, tela);
	}
	
	public void pintaSegmentoDeReta(int x1, int y1, int x2, int y2, Color c, int tela){

		Graphics graphics = telas.get(tela).getGraphics();
		graphics.setColor(c);

		graphics.drawLine(x1, y1, x2, y2);
	}
	
	public void pintaSegmentoDeReta(int x1, int y1, int x2, int y2, int tela){
		pintaSegmentoDeReta(x1, y1, x2, y2, Color.black, tela);
	}
	
	public void pintaSegmentoDeReta(iPoint ponto1, iPoint ponto2, Color c, int tela){
		pintaSegmentoDeReta(ponto1.getX(), ponto1.getY(), ponto2.getX(), ponto2.getY(), c, tela);
	}
	
	public void pintaSegmentoDeReta(iPoint ponto1, iPoint ponto2, int tela){
		pintaSegmentoDeReta(ponto1.getX(), ponto1.getY(), ponto2.getX(), ponto2.getY(), Color.black, tela);
	}
	
	public void pintaSegmentoDeReta(Ponto ponto1, Ponto ponto2, Color c, int tela){
		pintaSegmentoDeReta((int)ponto1.getX(), (int)ponto1.getY(), (int)ponto2.getX(), (int)ponto2.getY(), c, tela);
	}
	
	public void pintaSegmentoDeReta(Ponto ponto1, Ponto ponto2, int tela){
		pintaSegmentoDeReta((int)ponto1.getX(), (int)ponto1.getY(), (int)ponto2.getX(), (int)ponto2.getY(), Color.black, tela);
	}
	
	public void pintaSegmentoDeReta(LineSegment l, Color c, int tela){
		System.out.println("(" + (int)l.getP1().getX() +", " + (int)l.getP1().getY()
							+ " (" + (int)l.getP2().getX() +", " + (int)l.getP2().getY());
		pintaSegmentoDeReta((int)l.getP1().getX(), (int)l.getP2().getY(), (int)l.getP1().getX(), (int)l.getP2().getY(), c, tela);
	}
	
	public void pintaSegmentoDeRetaHorizontal(Ponto ponto1, Color c, int tela){
		pintaSegmentoDeReta(0, (int)ponto1.getY(), CANVAS_W, (int)ponto1.getY(), c, tela);
	}
	
	public void pintaSegmentoDeRetaHorizontal(int y, Color c, int tela){
		pintaSegmentoDeReta(0, y, CANVAS_W, y, c, tela);
	}
		
	public void limpaTela(int tela){
		telas.get(tela).getGraphics().clearRect(0, 0, CANVAS_W, CANVAS_H);
	}
	
	public void limpaTelas(){
		for(int i = 1; i < NUMERO_TELAS; i++){
			limpaTela(i);
		}
	}
	
	/**
	 * Copia a tela src em dst
	 * @param src
	 * @param dst
	 */
	public void copiaTela(int src, int dst){
		telas.get(dst).getGraphics().drawImage(telas.get(src), 0, 0, this);		
	}
	
	public void pintaPoligono(Polygon poly, int tela) {
		limpaTela(tela);
		Vertex v = poly.getHead();
		do {
			pintaPonto(v.getPoint(), tela);
			pintaSegmentoDeReta(v.getVPrev().getPoint(), v.getPoint(), tela);
			v = v.getVNext();
		} while(v != poly.getHead());
		pintaTela(tela);
	}

	public void setDrawingPolygon(boolean desenhandoPontos) {
		this.drawingPolygon = desenhandoPontos;
	}

	public boolean isDrawingPolygon() {
		return drawingPolygon;
	}

	public void setRunning(boolean running) {
		this.running = running;
	}

	public boolean isRunning() {
		return running;
	}

	public void setDrawingPoints(boolean drawingPoints) {
		this.drawingPoints = drawingPoints;
	}

	public boolean isDrawingPoints() {
		return drawingPoints;
	}
	
	public void pintaPontos(ArrayList<iPoint> points, Color c, int tela) {
		for(iPoint p : points){
			pintaPonto(p, c, tela);
		}
	}

	public void pintaPontos(ArrayList<iPoint> points, int tela) {
		pintaPontos(points, Color.black, tela);
	}
	
}